/*
 * VFAPI plugin - PNG reader. Written for tmgpenc - http://www.tmpgenc.com/
 *
 * vfapi.h is from the VFAPI package available from the tmpgenc home page.
 *   (I found it in the Japanese section; the English section is missing a few files).
 *
 * All the real work is done by libpng - http://www.libpng.org/pub/png/
 *  Of course, libpng depends on zlib - http://www.info-zip.org/pub/infozip/zlib/
 *
 * Written in VisualC 6, but I don't see why it wouldn't work with other compilers.
 *  _splitpath and _access are non-ANSI, but _splitpath at least is fairly common.
 *   (the functionality of _access is farily common, too. If you don't have it (or
 *    _stat), it is easily replaced with an fopen/fclose pair).
 *
 * Send bug reports to ix42@hotmail.com. Better yet, send patches instead.
 * A patch to verify the size of each image as we go would be nice.
 *
 * This code is public domain. Do what you like with it.
 *
 * Warranty: If this code breaks, you get to keep the pieces.
 *
 * Copyright 2001 Peter Harris
 *
 */


#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include "vfapi2.h"
#include <stdlib.h>
#include <io.h>
#include <png.h>
#include <malloc.h>

static void die(png_structp png_ptr, png_const_charp message)
{
    MessageBox(NULL,message,"Fatal Error",MB_OK|MB_ICONERROR);
    exit(EXIT_FAILURE);
}

static void whine(png_structp png_ptr, png_const_charp message)
{
    MessageBox(NULL,message,"Warning",MB_OK|MB_ICONINFORMATION);
}

HRESULT __stdcall vfGetPluginInfo( LPVF_PluginInfo lpPluginInfo )
{
    if (lpPluginInfo == NULL)
        return VF_ERROR;
    if (lpPluginInfo->dwSize != sizeof(VF_PluginInfo))
        return VF_ERROR;

    lpPluginInfo->dwAPIVersion = 1;
    lpPluginInfo->dwVersion = 1;
    lpPluginInfo->dwSupportStreamType = VF_STREAM_VIDEO;
    strcpy(lpPluginInfo->cPluginInfo,"Pete's PNG Reader");
    strcpy(lpPluginInfo->cFileType,"Portable Network Graphics (*.png)|*.png");

    return VF_OK;
}

typedef struct Stuff {
    char fn_format[_MAX_PATH+5];
    int first_frame;
    int total_frames;
} Stuff;

HRESULT __stdcall PNG_OpenFile(char *FileName, LPVF_FileHandle fh)
{
    char basename[_MAX_PATH];
    char ext[_MAX_EXT];
    FILE *F;
    unsigned char header[8];
    int i;
    int number_length = 0;
    Stuff *stuff;

    if (fh == NULL)
        return VF_ERROR;

    F = fopen(FileName, "rb");
    if (F == NULL)
        return VF_ERROR;

    i = fread(header, 1, 8, F);
    fclose(F);

    if (i != 8) //file too small
        return VF_ERROR;

    if (png_sig_cmp(header,0,8)) {
        //oops. Not a PNG.
        return VF_ERROR;
    }

    stuff = malloc(sizeof(Stuff));
    if (stuff == NULL)
        return VF_ERROR;

    _splitpath(FileName,NULL,NULL,NULL,ext);
    i = strlen(FileName) - strlen(ext);
    strncpy(basename, FileName, i);
    basename[i]='\0';

    i--;
    while (basename[i] >= '0' && basename[i] <= '9') {
        i--;
        number_length++;
    }
    i++;
    stuff->first_frame = atoi(basename + i);
    basename[i] = '\0';
    sprintf(stuff->fn_format,"%s%%0%dd%s",basename,number_length,ext);

    i = stuff->first_frame;
    do {
        sprintf(basename,stuff->fn_format,i++);
    } while (!_access(basename, 4)); // 4 == READ_ACCESS

    i--;
    stuff->total_frames = i - stuff->first_frame;

    *fh = (DWORD)stuff;
    return VF_OK;

}

HRESULT __stdcall PNG_CloseFile(VF_FileHandle fh)
{
    Stuff *stuff = (Stuff *)fh;
    if (fh == 0)
        return VF_OK;
    if (stuff == NULL)
        return VF_ERROR;

    free(stuff);

    return VF_OK;
}

HRESULT __stdcall PNG_GetFileInfo(VF_FileHandle fh, LPVF_FileInfo lpFileInfo)
{
    if (fh == 0)
        return VF_ERROR;
    if (lpFileInfo == NULL)
        return VF_ERROR;

    if (lpFileInfo->dwSize != sizeof(VF_FileInfo))
        return VF_ERROR;

    lpFileInfo->dwHasStreams = VF_STREAM_VIDEO;
    return VF_OK;
}

static int PNG_FillInInfo(Stuff *stuff, LPVF_StreamInfo_Video vi)
{
    png_structp read;
    png_infop info;
    png_byte ct;

    char fn[_MAX_PATH];
    FILE *F;
    sprintf(fn,stuff->fn_format,stuff->first_frame);
    
    F = fopen(fn,"rb");
    if (F == NULL)
        return 0;

    read = png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,die,whine);
    if (read == NULL) {
        fclose(F);
        return 0;
    }
    info = png_create_info_struct(read);
    if (info == NULL) {
        fclose(F);
        png_destroy_read_struct(&read,NULL,NULL);
        return 0;
    }

    png_init_io(read, F);
    png_read_info(read, info);

    ct = png_get_color_type(read,info);
    if (ct != PNG_COLOR_TYPE_RGB &&        //don't support greys. Yet, anyway.
        ct != PNG_COLOR_TYPE_RGB_ALPHA &&  //TRANSFORM_STRIP handles this.
        ct != PNG_COLOR_TYPE_PALETTE)      //TRANSFORM_EXPAND handles this.
        return 0;

    vi->dwWidth = png_get_image_width(read,info);
    vi->dwHeight = png_get_image_height(read,info);
    vi->dwBitCount = 24; //for now, we'll force libpng to convert to 24-bit for us.
                         //does tmpgenc accept any other depths?

    png_destroy_read_struct(&read, &info, NULL);
    fclose(F);
    return -1; //success!
}

HRESULT __stdcall PNG_GetStreamInfo(VF_FileHandle fh, DWORD dwStream, void *stream_info)
{
    LPVF_StreamInfo_Video info = (LPVF_StreamInfo_Video)stream_info;
    Stuff *stuff = (Stuff *)fh;
    if (fh == 0)
        return VF_ERROR;
    if (stuff == NULL)
        return VF_ERROR;

    if (dwStream != VF_STREAM_VIDEO)
        return VF_ERROR;

    if (info == NULL)
        return VF_ERROR;

    if (info->dwSize != sizeof(VF_StreamInfo_Video))
        return VF_ERROR;

    info->dwLength = stuff->total_frames;
    info->dwScale = 0;
    info->dwRate = 0;

    if (!PNG_FillInInfo(stuff, info))
        return VF_ERROR;

    return VF_OK;
}

HRESULT __stdcall PNG_ReadData(VF_FileHandle fh, DWORD dwStream, void *data)
{
    png_structp read;
    png_infop png_info;
    png_bytep *rows;

    unsigned char *dst;
    int width, height;
    int y;
    char fn[_MAX_PATH];
    FILE *F;

    LPVF_ReadData_Video info = (LPVF_ReadData_Video)data;
    Stuff *stuff = (Stuff *)fh;

    if (fh == 0)
        return VF_ERROR;
    if (stuff == NULL)
        return VF_ERROR;

    if (dwStream != VF_STREAM_VIDEO)
        return VF_ERROR;

    if (info == NULL)
        return VF_ERROR;
    if (info->dwSize != sizeof(VF_ReadData_Video))
        return VF_ERROR;
    if (info->dwFrameNumber < 0 || info->dwFrameNumber >= stuff->total_frames)
        return VF_ERROR;
    if (info->lpData == NULL)
        return VF_ERROR;
    if (info->lPitch == 0)
        return VF_ERROR;

    sprintf(fn,stuff->fn_format,stuff->first_frame + info->dwFrameNumber);
    
    F = fopen(fn,"rb");
    if (F == NULL)
        return 0;

    read = png_create_read_struct(PNG_LIBPNG_VER_STRING,NULL,die,whine);
    if (read == NULL) {
        fclose(F);
        return 0;
    }
    png_info = png_create_info_struct(read);
    if (png_info == NULL) {
        fclose(F);
        png_destroy_read_struct(&read,NULL,NULL);
        return 0;
    }

    //TODO: check width and height here.

    png_init_io(read, F);
    png_read_png(read, png_info, 
        PNG_TRANSFORM_STRIP_16 |
        PNG_TRANSFORM_STRIP_ALPHA |
        PNG_TRANSFORM_EXPAND |
        PNG_TRANSFORM_BGR        //they want BGR. Go figure.
        , NULL);

    rows = png_get_rows(read, png_info);
    width = png_get_image_width(read,png_info);
    height = png_get_image_height(read,png_info);
    dst = info->lpData;
    for (y=0; y<height; y++) {
        memcpy(dst, rows[y], width * 3);
        dst += info->lPitch;
    }

    png_destroy_read_struct(&read, &png_info, NULL);
    fclose(F);

    return VF_OK;
}

HRESULT __stdcall vfGetPluginFunc( LPVF_PluginFunc lpPluginFunc )
{
    if (lpPluginFunc == NULL)
        return VF_ERROR;
    if (lpPluginFunc->dwSize != sizeof(VF_PluginFunc))
        return VF_ERROR;

    lpPluginFunc->CloseFile = PNG_CloseFile;
    lpPluginFunc->GetFileInfo = PNG_GetFileInfo;
    lpPluginFunc->GetStreamInfo = PNG_GetStreamInfo;
    lpPluginFunc->OpenFile = PNG_OpenFile;
    lpPluginFunc->ReadData = PNG_ReadData;

    return VF_OK;
}
